#include "AccImage.h"

AccImage::AccImage()
{
    __rows = 0;
    __cols = 0;
}

AccImage::AccImage(const RgbImage& rgb)
{
    float rgbacc[3][3] = {
    { 1.048448571846226,  0.993060609426868,  0.859793008692159},
    { 0.416525530270566, -0.089902211368311, -0.250822792497467},
    {-0.480843560763957, -0.418987267899313,  3.560717387060904}};
                
    __rows = rgb.rows();
    __cols = rgb.cols();
    __A = Array<float>(__rows, __cols);
    __C1 = Array<float>(__rows, __cols);
    __C2 = Array<float>(__rows, __cols);
    
    for (size_t r = 0; r < rgb.rows(); r++)
    {
        for (size_t c = 0; c < rgb.cols(); c++)
        {
            // Gama correction to RGB value.
            float L = 0.13 + 18.18  * powf((float)rgb.R()(r, c) / 
                                            255.0, 2.1649);
            float M = 0.13 + 58.542 * powf((float)rgb.G()(r, c) / 
                                            255.0, 2.195);
            float S = 0.13 + 9.2465 * powf((float)rgb.B()(r, c) /
                                            255.0, 2.1812);
            
            // Krauskopf
            // |  A  |   |    1    1    0    |   | L |
            // | Cr1 | = |    1   -1    0    | * | M |
            // | Cr2 |   |  -0.5 -0.5   1    |   | S |
            
            __A(r, c) = (rgbacc[0][0] * L) + 
                        (rgbacc[0][1] * M) + 
                        (rgbacc[0][2] * S);
            
            __C1(r, c) = (rgbacc[1][0] * L) + 
                         (rgbacc[1][1] * M) + 
                         (rgbacc[1][2] * S);
            
            __C2(r, c) = (rgbacc[2][0] * L) +
                         (rgbacc[2][1] * M) + 
                         (rgbacc[2][2] * S);
        }
    }    
}

AccImage::~AccImage()
{
}

size_t
AccImage::rows() const
{
    return __rows;
}

size_t
AccImage::cols() const
{
    return __cols;
}

const Array<float>&
AccImage::A() const
{
    return __A;
}

const Array<float>&
AccImage::C1() const
{
    return __C1;
}

const Array<float>&
AccImage::C2() const
{
    return __C2;
}

std::ostream& 
operator<<(std::ostream& os, const AccImage& img)
{
    os << "A" << std::endl;
    os << img.__A << std::endl;
    os << "C1" << std::endl;
    os << img.__C1 << std::endl;
    os << "C2" << std::endl;
    os << img.__C2 << std::endl;
    
    return os;
}