#include "ArgumentLoader.h"

ArgumentLoader::ArgumentLoader(int argc, char *argv[])
{


   errors.clear();
   argumentMap.clear();
   
   argumentStr  = "";

   /* Not enough arguments */
   if (argc < 2) return;

   for(int i=1; i < argc; i++)
   {

      argumentStr += argv[i];
      if (i != (argc-1)) argumentStr += " ";

   }


};


ArgumentLoader::~ArgumentLoader()
{

   errors.clear();
   argumentMap.clear();

};


bool ArgumentLoader::addPattern(string argtok, int subargcnt, bool comp)
{

   if (!argumentExists(argtok))
   {

      Argument arg = Argument(argtok, subargcnt, comp);
      argumentMap[argtok] = arg;
      return true;

   }
   else
    return false;

};


bool ArgumentLoader::argumentExists(string argtok)
{

  return (argumentMap.find(argtok.c_str()) != argumentMap.end());

};


string ArgumentLoader::getArgumentString()
{

   return argumentStr;

};


int ArgumentLoader::getArgumentCount()
{

   return argumentMap.size();

};


Argument ArgumentLoader::getArgument(string argtok)
{

   return argumentMap[argtok];

};


bool ArgumentLoader::parseArguments()
{

   bool result = true;

   StringTokenizer strtok = StringTokenizer(argumentStr,"-");

   while (strtok.hasMoreTokens())
   {

      string currentBlock = strtok.nextToken();

      StringTokenizer cbtok = StringTokenizer(currentBlock," ");

      string argument = cbtok.nextToken();

      if (argumentExists(argument))
      {

         Argument arg = argumentMap[argument];

         arg.setInitiated();

         //if ((cbtok.countTokens() == arg.getexpfectedSubArgCount()) || (arg.getSubArgCount() < 0))
         if ((cbtok.countTokens() == arg.getexpfectedSubArgCount()) || (arg.getexpfectedSubArgCount() < 0))
         {


            while(cbtok.hasMoreTokens())
            {

               arg.addSubArg(cbtok.nextToken());

            }

            /* Update argument */
            argumentMap[argument] = arg;

         }
         else
         {

            stringstream  errStr;
            errStr << "Incorrect number of sub arguments! Argument: [" << arg.getArgToken() << "]   Token Count: [" << cbtok.countTokens() << "]  Sub-Arg Count: [" << arg.getexpfectedSubArgCount() << "] ";
            errors.push_back(errStr.str());
            result = false;
            break;

         }

      }
      else
      {

         stringstream  errStr;
         errStr << "Argument Pattern DOES NOT EXIST within loader [" << argument << "]";
         errors.push_back(errStr.str());
         result = false;
         break;


      }




   }



   /*
      Check to see if any compolsary argument patterns
      have not been provided.
   */
   for (ArgumentMap::iterator it = argumentMap.begin(); it != argumentMap.end(); ++it)
   {

      Argument arg = it->second;

      if (arg.isCompulsory() && (! arg.isInitiated()))
      {
 
         stringstream  errStr;
         errStr << "Compulsory Argument Pattern DOES NOT EXIST within loader [" << arg.getArgToken() << "]";
         errors.push_back(errStr.str());
         result = false;
         
      }

   }


   return result;

};


StringList ArgumentLoader::getErrors()
{
	
   return errors;	
   
};

