#include <cmath>
#include <vector>

#include "CsfFilter.h"

CsfFilter::CsfFilter(size_t rows, size_t cols)
: __rows(rows), __cols(cols)
{       
    size_t r;
    size_t c;
    size_t offset0w = cols / 2 + 1;
    size_t offset0h = rows / 2 + 1;
    float delta0w = 1.0 / (float)cols;
    float delta0h = 1.0 / (float)rows;
    std::vector<float> fl(offset0h);
    std::vector<float> fc(offset0w);
    std::vector<std::vector<float> > Fc(offset0h);
    std::vector<std::vector<float> > Fl(offset0h);
    for (size_t i = 0; i < offset0h; i ++)
    {
        Fc[i].resize(offset0w);
        Fl[i].resize(offset0w);
    }
    /*
    float Fc[offset0h][offset0w];
    float Fl[offset0h][offset0w];
    */
    float i2 = (float)((float)rows * (float)cols / (float)3600.0);
#ifdef DEBUG
    std::cout << "I2" << i2 << std::endl;
#endif
    
    __A = Array<float>(rows, cols);
    __C1 = Array<float>(rows, cols);
    __C2 = Array<float>(rows, cols);
    for (c = 1; c <= offset0w; c ++)
        fc[c - 1] = (float)c * delta0h;
    
    for (r = 1; r <= offset0h; r++) 
        fl[r - 1] = (float)r * delta0w;
    
    for (r = 0; r < offset0h; r++)
    {
        for (c = 0; c < offset0w; c++)
        {
            Fc[r][c] = fc[r];
            Fl[r][c] = fl[c];
        }
    }

#ifdef DEBUG
    std::cout << "Fc" << std::endl;
    for (r = 0; r < offset0h; r++)
    {
        for (c = 0; c < offset0w; c++)
        {
            std::cout << Fc[r][c] << "\t";
        }
        std::cout << std::endl;
    }

    std::cout << "Fl" << std::endl;
    for (r = 0; r < offset0h; r++)
    {
        for (c = 0; c < offset0w; c++)
        {
            std::cout << Fl[r][c] << "\t";
        }
        std::cout << std::endl;
    }
#endif
    
    Array<float> csf_a(offset0h, offset0w);
    Array<float> csf_c1(offset0h, offset0w);
    Array<float> csf_c2(offset0h, offset0w);

    float max = 0;
    for (r = 0; r < offset0h; r++)
    {
        for (c = 0; c < offset0w; c++)
        {
            float w1;
            float w2;
            float cs1;
            float cs2;
            float csf1;
            float csf2;
            float csf3;
            
            w1 = 60.0 * sqrtf(Fl[r][c] * Fl[r][c] + Fc[r][c] * Fc[r][c]);
#ifdef DEBUG
            std::cout << "W1 " << w1 << std::endl;
#endif
            
            w2 = w1 / ((0.856 * powf(2.5, 0.14)) /** re*/ * (0.11 * cosf(4.0 * 
                (atan2f(Fl[r][c], Fc[r][c]))) + 0.89));
#ifdef DEBUG
            std::cout << "W2 " << w2 << std::endl;
#endif            
            /* A */
            cs1 = powf(powf(3.23 * powf(i2 * w1 * w1, -0.3), 5) + 1, -0.2);
#ifdef DEBUG
            std::cout << "CS1 " << cs1 << std::endl;
#endif                        
            cs2 = 0.801 * powf((1.7 / 100.0), -0.2) * 0.9 * cs1 * w1 * 
                expf(-(0.3 * powf((101.0 / 100.0), 0.15)) * 0.9 * w1);
#ifdef DEBUG
            std::cout << "CS2 " << cs2 << std::endl;
#endif                        
            csf1 = cs2 * powf(1.0 + 0.06 * expf((0.3 * powf((101.0 / 100.0), 
                0.15)) * 0.9 * w1), 0.5);
#ifdef DEBUG
            std::cout << "CSF1 " << csf1 << std::endl;
#endif                                    
            
            cs1 = powf(powf(3.23 * powf(i2 * w2 * w2, -0.3), 5.0) + 1.0, -0.2);
#ifdef DEBUG
            std::cout << "CS1 " << cs1 << std::endl;
#endif                                    
            cs2 = 0.801 * powf((1.7 / 100.0), -0.2) * 0.9 * cs1 * w2 * 
                expf(-(0.3 * powf((101.0 / 100.0), 0.15)) * 0.9 * w2);
#ifdef DEBUG
            std::cout << "CS2 " << cs2 << std::endl;
#endif                        
            csf2 = cs2 * powf(1.0 + 0.06 * expf((0.3 * powf((101.0 / 100.0), 
                0.15)) * 0.9 * w2), 0.5);
#ifdef DEBUG
            std::cout << "CSF2 " << csf2 << std::endl;
#endif                                    
            
            csf3 = csf1 < csf2 ? csf1 : csf2;
#ifdef DEBUG
            std::cout << "CSF3 " << csf3 << std::endl;
#endif                                    
            
            max = max > csf3 ? max : csf3;
            
            csf_a(r, c) = csf3;
            
            /* CR1 */
            csf_c1(r, c) = 1.0 / (1.0 + powf(w1 / 5.52, 1.72)) * 
                (1 - 0.27 * sinf(2 * atan2f(Fl[r][c], Fc[r][c])));
            
            /* CR2 */
            csf_c2(r, c) = 1.0 / (1.0 + powf(w1 / 4.123, 1.6454)) * 
                (1.0 - 0.24 * sinf(2 * atan2f(Fl[r][c], Fc[r][c])));
        }
    }
    
    csf_a /= max;
    
    // csf(offset0h-1:h,offset0w-1:l) = csf3;
    for (r = offset0h - 2; r < rows; r ++)
    {
        for (c = offset0w - 2; c < cols; c ++)
        {
            __A(r, c) = csf_a(r - offset0h + 2, c - offset0w + 2);
            __C1(r, c) = csf_c1(r - offset0h + 2, c - offset0w + 2);
            __C2(r, c) = csf_c2(r - offset0h + 2, c - offset0w + 2);
        }
    }
    Array<float> flip_a(csf_a);
    flip_a.flipud();
    Array<float> flip_cr1(csf_c1);
    flip_cr1.flipud();
    Array<float> flip_cr2(csf_c2);
    flip_cr2.flipud();
    
    // csf(1:offset0h,offset0w-1:l) = flipud(csf3);
    for (r = 0; r < offset0h ; r++)
    {
        for (c = offset0w - 2; c < cols; c++)
        {
            __A(r, c) = flip_a(r, c - offset0w + 2);
            __C1(r, c) = flip_cr1(r, c - offset0w + 2);
            __C2(r, c) = flip_cr2(r, c - offset0w + 2);
        }
    }

    flip_a.fliplr();
    flip_cr1.fliplr();
    flip_cr2.fliplr();
    
    // csf(1:offset0h,1:offset0w) = fliplr(flipud(csf3));
    for (r = 0; r < offset0h ; r++)
    {
        for (c = 0; c < offset0w; c++)
        {
            __A(r, c) = flip_a(r, c);
            __C1(r, c) = flip_cr1(r, c);
            __C2(r, c) = flip_cr2(r, c);
        }
    }

    // csf(offset0h-1:h,1:offset0w) = fliplr(csf3);
    csf_a.fliplr();
    csf_c1.fliplr();
    csf_c2.fliplr();
    for (r = offset0h - 2; r < __rows; r++)
    {
        for (c = 0; c < offset0w; c++)
        {
            __A(r, c) = csf_a(r - offset0h + 2, c);
            __C1(r, c) = csf_c1(r - offset0h + 2, c);
            __C2(r, c) = csf_c2(r - offset0h + 2, c);
        }
    }
    
    __A *= 250.0;
    __C1 *= 33.0;
    __C2 *= 5.0;
}

const Array<float>& 
CsfFilter::A()
{
    return __A;
}

const Array<float>&
CsfFilter::C1()
{
    return __C1;
}

const Array<float>& 
CsfFilter::C2()
{
    return __C2;
}

std::ostream& 
operator<<(std::ostream& os, const CsfFilter& csf)
{
    os << "A" << std::endl;
    os << csf.__A << std::endl;
    os << "C1" << std::endl;
    os << csf.__C1 << std::endl;
    os << "C2" << std::endl;
    os << csf.__C2 << std::endl;
    
    return os;
}