#include <vector>

#include "DcpFilter.h"

#define PI acos(-1.0)
//#define PI 3.1415

DcpFilter::DcpFilter(size_t rows, size_t cols)
: __rows(rows), __cols(cols)
{
    size_t offset0w = cols / 2 + 1;
    size_t offset0h = rows / 2 + 1;
    float delta0w = 1.0 / (float)cols;
    float delta0h = 1.0 / (float)rows;

    for (int i = 0; i < 27; i++)
        __dcp[i] = Array<float>(__rows, __cols);

    // [Fl,Fc] = meshgrid(fl,fc)
    /* gcc is ok but...
    float fl[__rows];
    float fc[__cols];
    float Fc[__rows][__cols];
    float Fl[__rows][__cols];
    */
    std::vector<float> fl(__rows);
    std::vector<float> fc(__cols);
    std::vector<std::vector<float> > Fc(__rows);
    std::vector<std::vector<float> > Fl(__rows);
    for (size_t i = 0; i < __rows; i ++)
    {
        Fc[i].resize(__cols);
        Fl[i].resize(__cols);
    }
    
    for (size_t c = 1; c <= __cols; c ++)
        fc[c - 1] = (float)(((float)c - (float)offset0w) * (float)delta0w);
    
    for (size_t r = 1; r <= __rows; r++) 
        fl[r - 1] = (float)(((float)r - (float)offset0h) * (float)delta0h);
    
    for (size_t r = 0; r < __rows; r++)
    {
        for (size_t c = 0; c < __cols; c++)
        {
            // XXX : ??????????????
            Fc[r][c] = fl[r];
            Fl[r][c] = fc[c];
        }
    }
 
    // Mesa init
    Array<float> w1(__rows, __cols);
    Array<float> base(__rows, __cols);
    Array<float> mesa2(__rows, __cols);
    Array<float> mesa3(__rows, __cols);
    Array<float> mesa4(__rows, __cols);

    for (size_t r = 0; r < __rows; r++)
        for (size_t c = 0; c < __cols; c++)
            w1(r, c) = (float)(sqrtf(Fl[r][c] * Fl[r][c] + 
                                    Fc[r][c] * Fc[r][c]));
    
    // Fan init
    Array<float> teta(__rows, __cols);
    Array<float> ateta(__rows, __cols);
    Array<float> fan_45_1(__rows, __cols);
    Array<float> fan_45_2(__rows, __cols);
    Array<float> fan_45_3(__rows, __cols);
    Array<float> fan_45_4(__rows, __cols);
    Array<float> fan_30_1(__rows, __cols);
    Array<float> fan_30_2(__rows, __cols);
    Array<float> fan_30_3(__rows, __cols);
    Array<float> fan_30_4(__rows, __cols);
    Array<float> fan_30_5(__rows, __cols);
    Array<float> fan_30_6(__rows, __cols);
    
    for (size_t r = 0; r < __rows; r++)
    {
        for (size_t c = 0; c < __cols; c++)
        {
            float t = (float)(- (atan2(Fl[r][c], Fc[r][c])));
            teta(r, c) = t; 
            ateta(r, c) = fabs(t);
        }
    }

    //
    // DCP A;
    //
    __mesa(base, w1, 1.5 / 60.0, 2.6 / 60.0);
    __mesa(mesa2, w1, 5.7 / 60.0, 4.0 / 60.0);
    __mesa(mesa3, w1, 14.4 / 60.0, 8.0 / 60.0);
    __mesa(mesa4, w1, 28.8 / 60.0, 15.0 / 60.0);

    __fan(fan_45_1, teta, PI/4.0, 20.0/180.0*PI, 0.0);
    __fan(fan_45_2, teta, PI/4.0, 20.0/180.0*PI, PI/4.0);
    __fan(fan_45_3, ateta, PI/4.0, 20.0/180.0*PI, -PI/2.0);
    __fan(fan_45_4, teta, PI/4.0, 20.0/180.0*PI, -PI/4.0);

    __dcp[0] = base;

    Array<float> dom(mesa2 - base);
    __dcp[1] = dom * fan_45_1;
    __dcp[2] = dom * fan_45_2;
    __dcp[3] = dom * fan_45_3;
    __dcp[4] = dom * fan_45_4;

    __fan(fan_30_1, teta, PI/6.0, 10.0/180.0*PI, 0.0);
    __fan(fan_30_2, teta, PI/6.0, 10.0/180.0*PI, PI/6.0);
    __fan(fan_30_3, teta, PI/6.0, 10.0/180.0*PI, PI/3.0);
    __fan(fan_30_4, ateta, PI/6.0, 10.0/180.0*PI, -PI/2.0);
    __fan(fan_30_5, teta, PI/6.0, 10.0/180.0*PI, -PI/3.0);
    __fan(fan_30_6, teta, PI/6.0, 10.0/180.0*PI, -PI/6.0);

    dom = mesa3 - mesa2;
    __dcp[5] = dom * fan_30_1;
    __dcp[6] = dom * fan_30_2;
    __dcp[7] = dom * fan_30_3;
    __dcp[8] = dom * fan_30_4;
    __dcp[9] = dom * fan_30_5;
    __dcp[10] = dom * fan_30_6;
    
    __fan(fan_30_1, teta, PI/6.0, 8.0/180.0*PI, 0.0);
    __fan(fan_30_2, teta, PI/6.0, 8.0/180.0*PI, PI/6.0);
    __fan(fan_30_3, teta, PI/6.0, 8.0/180.0*PI, PI/3.0);
    __fan(fan_30_4, ateta, PI/6.0, 8.0/180.0*PI, -PI/2.0);
    __fan(fan_30_5, teta, PI/6.0, 8.0/180.0*PI, -PI/3.0);
    __fan(fan_30_6, teta, PI/6.0, 8.0/180.0*PI, -PI/6.0);

    dom = mesa4 - mesa3;
    __dcp[11] = dom * fan_30_1;
    __dcp[12] = dom * fan_30_2;
    __dcp[13] = dom * fan_30_3;
    __dcp[14] = dom * fan_30_4;
    __dcp[15] = dom * fan_30_5;
    __dcp[16] = dom * fan_30_6;

    //
    // DCP CR1
    //
    __mesa(base, w1, 1.5/60.0, 2.8/60.0);
    __mesa(mesa2, w1, 5.7/60.0, 4.8/60.0);

    __fan(fan_45_1, teta, PI/4.0, 45.0/180.0*PI, 0.0);
    __fan(fan_45_2, teta, PI/4.0, 45.0/180.0*PI, PI/4.0);
    __fan(fan_45_3, ateta, PI/4.0, 45.0/180.0*PI, -PI/2.0);
    __fan(fan_45_4, teta, PI/4.0, 45.0/180.0*PI, -PI/4.0);

    __dcp[17] = base;

    dom = mesa2 - base;
    __dcp[18] = dom * fan_45_1;
    __dcp[19] = dom * fan_45_2;
    __dcp[20] = dom * fan_45_3;
    __dcp[21] = dom * fan_45_4;
    
    //
    // DCP CR2
    //
    __mesa(base, w1, 1.5/60.0, 2.95/60.0);
    __mesa(mesa2, w1, 5.7/60.0, 5.8/60.0);

    __dcp[22] = base;
    
    dom = mesa2 - base;
    __dcp[23] = dom * fan_45_1;
    __dcp[24] = dom * fan_45_2;
    __dcp[25] = dom * fan_45_3;
    __dcp[26] = dom * fan_45_4;    
}

const Array<float>&
DcpFilter::operator[](size_t i) const {return __dcp[i];}

const Array<float>& 
DcpFilter::ABase() const {return __dcp[0];}

const Array<float>& 
DcpFilter::ASub21() const {return __dcp[1];}

const Array<float>&
DcpFilter::ASub22() const {return __dcp[2];}

const Array<float>&
DcpFilter::ASub23() const {return __dcp[3];}

const Array<float>& 
DcpFilter::ASub24() const {return __dcp[4];}

const Array<float>&
DcpFilter::ASub31() const {return __dcp[5];}

const Array<float>& 
DcpFilter::ASub32() const {return __dcp[6];}

const Array<float>& 
DcpFilter::ASub33() const {return __dcp[7];}

const Array<float>& 
DcpFilter::ASub34() const {return __dcp[8];}

const Array<float>&
DcpFilter::ASub35() const {return __dcp[9];}

const Array<float>&
DcpFilter::ASub36() const {return __dcp[10];}

const Array<float>& 
DcpFilter::ASub41() const {return __dcp[11];}

const Array<float>&
DcpFilter::ASub42() const {return __dcp[12];}

const Array<float>&
DcpFilter::ASub43() const {return __dcp[13];}

const Array<float>&
DcpFilter::ASub44() const {return __dcp[14];}

const Array<float>&
DcpFilter::ASub45() const {return __dcp[15];}

const Array<float>& 
DcpFilter::ASub46() const {return __dcp[16];}

const Array<float>&
DcpFilter::C1Base() const {return __dcp[17];}

const Array<float>&
DcpFilter::C1Sub21() const {return __dcp[18];}

const Array<float>&
DcpFilter::C1Sub22() const {return __dcp[19];}

const Array<float>&
DcpFilter::C1Sub23() const {return __dcp[20];}

const Array<float>& 
DcpFilter::C1Sub24() const {return __dcp[21];}

const Array<float>&
DcpFilter::C2Base() const {return __dcp[22];}

const Array<float>& 
DcpFilter::C2Sub21() const {return __dcp[23];}

const Array<float>& 
DcpFilter::C2Sub22() const {return __dcp[24];}

const Array<float>& 
DcpFilter::C2Sub23() const {return __dcp[25];}

const Array<float>& 
DcpFilter::C2Sub24() const {return __dcp[26];}

void
DcpFilter::__mesa(Array<float>& m, Array<float>& w1, float fc, float lt)
{
    for (size_t r = 0;r < __rows; r++)
    {
        for (size_t c = 0; c < __cols; c++)
        {
            //m(r, c) = 0; Done by Array ctor.
            if (w1(r, c) < ((float)(fc + lt / (float)2.0)))
                m(r, c) = (float)(0.5 * (1.0 + cos(PI * (w1(r, c) - fc + lt  
                                                       / 2.0) / lt)));
            
            if (w1(r, c) < ((float)(fc - lt / (float)2.0 )))
                m(r, c) = 1;
        }
    }    
}

void
DcpFilter::__fan(Array<float>& f, Array<float>& teta, float tc, float tw, 
                 float td)
{
    for (size_t r = 0; r < __rows; r++)
    {
        for (size_t c = 0; c < __cols; c++)
        {
            float interm;
            interm = (float)(fabs(teta(r, c) + (PI / 2.0) + td));
            if (interm < ((tc / 2.0) + (tw / 2.0)))
                f(r, c) = (float)(0.5 * (1.0 + cos(PI * fabs(interm - (tc / 2.0) + (tw / 2.0)) / tw)));
            if (interm < ((tc / 2.0) - (tw / 2.0)))
                f(r, c) = 1.0;
            
            interm =(float)(fabs(teta(r, c) - (PI / 2.0) + td));
            if (interm < ((tc / 2.0) + (tw / 2.0)))
                f(r, c) = (float)(0.5 * ( 1.0 + cos(PI * fabs(interm - tc / 2.0 + tw / 2.0 ) / (tw))));
            if (interm < ((tc / 2.0) - (tw / 2.0)))
                f(r, c) = 1.0;
        }
    }
}

std::ostream&
operator<<(std::ostream& os, const DcpFilter& dcp)
{
    os << "A Base" << std::endl;
    os << dcp.ABase() << std::endl;
    os << "A 21" << std::endl;
    os << dcp.ASub21() << std::endl;
    os << "A 22" << std::endl;
    os << dcp.ASub22() << std::endl;
    os << "A 23" << std::endl;
    os << dcp.ASub23() << std::endl;
    os << "A 24" << std::endl;
    os << dcp.ASub24() << std::endl;
    os << "A 31" << std::endl;
    os << dcp.ASub31() << std::endl;
    os << "A 32" << std::endl;
    os << dcp.ASub32() << std::endl;
    os << "A 33" << std::endl;
    os << dcp.ASub33() << std::endl;
    os << "A 34" << std::endl;
    os << dcp.ASub34() << std::endl;
    os << "A 35" << std::endl;
    os << dcp.ASub35() << std::endl;
    os << "A 36" << std::endl;
    os << dcp.ASub36() << std::endl;
    os << "A 41" << std::endl;
    os << dcp.ASub41() << std::endl;
    os << "A 42" << std::endl;
    os << dcp.ASub42() << std::endl;
    os << "A 43" << std::endl;
    os << dcp.ASub43() << std::endl;
    os << "A 44" << std::endl;
    os << dcp.ASub44() << std::endl;
    os << "A 45" << std::endl;
    os << dcp.ASub45() << std::endl;
    os << "A 46" << std::endl;
    os << dcp.ASub46() << std::endl;
    os << "C1 Base" << std::endl;
    os << dcp.C1Base() << std::endl;
    os << "C1 21" << std::endl;
    os << dcp.C1Sub21() << std::endl;
    os << "C1 22" << std::endl;
    os << dcp.C1Sub22() << std::endl;
    os << "C1 23" << std::endl;
    os << dcp.C1Sub23() << std::endl;
    os << "C1 24" << std::endl;
    os << dcp.C1Sub24() << std::endl;
    os << "C2 Base" << std::endl;
    os << dcp.C2Base() << std::endl;
    os << "C2 21" << std::endl;
    os << dcp.C2Sub21() << std::endl;
    os << "C2 22" << std::endl;
    os << dcp.C2Sub22() << std::endl;
    os << "C2 23" << std::endl;
    os << dcp.C2Sub23() << std::endl;
    os << "C2 24" << std::endl;
    os << dcp.C2Sub24() << std::endl;
    /*
    for (size_t i = 0; i < 27; i++)
        os << dcp.__dcp[i] << std::endl << std::endl;
    */
    return os;
}
