#include <complex>

#include "Komparator2.h"
#include "IntraATresholdBuilder.h"
#include "IntraCTresholdBuilder.h"
#include "InterATresholdBuilder.h"
#include "InterBTresholdBuilder.h"
#include "CannyBuilder.h"
#include "fftw++.h"

#undef DEBUG
#ifdef DEBUG
#include <sstream>
#include <fstream>
#endif


// COEFs
// Intra mask A
#define COEF1 0.7
#define COEF2 1.0    
#define COEF3 0.85

// Inter A mask coef
#define COEF4 1.89
#define COEF5 0.03   
#define COEF6 2.0
#define COEF7 1.7
#define COEF8 0.7    
#define COEF9 0.22   
#define COEF10 0.089
#define COEF11 0.0005
#define COEF12 0.1 
#define COEF13 2.0
#define COEF14 1.0
#define COEF15 0.0016
#define COEF16 0.141
#define COEF17 0.002
#define COEF18 0.238
#define COEF19 0.019
#define COEF20 0.0008
#define COEF21 0.064

// Intra Cr1 mask coef  
#define COEF22 0.45
#define COEF23 0.06
#define COEF24 1.22
#define COEF25 0.48
#define COEF26 0.03
#define COEF27 0.82

// Inter Cr1 mask coef
#define COEF28 -0.287
#define COEF29 0.053
#define COEF30 0.241
#define COEF31 0.09
#define COEF32 0.1
#define COEF33 0.82
#define COEF34 -0.057
#define COEF35 0.031
#define COEF36 0.258
#define COEF37 0.0232
#define COEF38 0.0013
#define COEF39 0.1096
#define COEF40 1.16
#define COEF41 0.16
#define COEF42 0.048

// Intra Cr2 mask coef
#define COEF43 0.72
#define COEF44 0.22
#define COEF45 2.78
#define COEF46 0.13
#define COEF47 0.05
#define COEF48 0.42

// Inter Cr2 mask coef
#define COEF49 0.136
#define COEF50 0.004
#define COEF51 0.196
#define COEF52 1.8
#define COEF53 0.8
#define COEF54 0.02

// Helpers functions.

int module(int i, int j);
Array<float> blkproc(Array<float>& in);
Array<int> edges(const AccImage& acc);
Array<int> erode_dilate_v(Array<int>& in);
Array<int> erode_dilate_h(Array<int>& in);
Array<int> dilate_hv(Array<int>& in);
Array<std::complex<float> > fft(const Array<float>& in);
Array<float> ifft(const Array<std::complex<float> >& in);
template<class T>
void to_file(const char* fname, const Array<T>& array);

// XXX : move to init.
Komparator2::Komparator2(const RgbImage& original, float alpha, float beta, 
                        float co, float a, float b, 
                        bool maskIntraA, bool maskInterA,
                        bool maskIntraC1, bool maskInterC1,
                        bool maskIntraC2, bool maskInterC2)
:__alpha(alpha), __beta(beta), __co(co), __a(a), __b(b),
__maskIntraA(maskIntraA), __maskInterA(maskInterA), __maskIntraC1(maskIntraC1),
__maskInterC1(maskInterC1),__maskIntraC2(maskIntraC2),__maskInterC2(maskInterC2)
{
    __csf = CsfFilter(original.rows(), original.cols());
    __dcp = DcpFilter(original.rows(), original.cols());

    AccImage acc(original);
    
    for (size_t i = 0; i < 27; i++)
        __ndcpor[i] = Array<float>(original.rows(), original.cols());    

    __decompose(acc, __ndcpor);
    __eor = edges(acc);
    
#ifdef DEBUG
    to_file("orig_r", original.R());
    to_file("orig_g", original.G());
    to_file("orig_b", original.B());
    to_file("orig_a", acc.A());
    to_file("orig_c1", acc.C1());
    to_file("orig_c2", acc.C2());
    to_file("csf_filter_a", __csf.A());
    to_file("csf_filter_c1", __csf.C1());
    to_file("csf_filter_c2", __csf.C2());
    for (size_t i = 0; i < 27; i++)
    {
        std::stringstream ss;
        ss << "dcp_filter_" << i; 
        to_file(ss.str().c_str(), __dcp[i]);
    }
    for (size_t i = 0; i < 27; i++)
    {
        std::stringstream ss;
        ss << "dcp_o_" << i; 
        to_file(ss.str().c_str(), __ndcpor[i]);
    }
    to_file("eor", __eor);
#endif
    
}

Comparison 
Komparator2::operator()(const RgbImage& degraded)
{
    Array<float> dndcp[27];
    for (size_t i = 0; i < 27; i++)
        dndcp[i] = Array<float>(degraded.rows(), degraded.cols());
    
    AccImage acc(degraded);
    
    __decompose(acc, dndcp);
#ifdef DEBUG
    for (size_t i = 0; i < 27; i++)
    {
        std::stringstream ss;
        ss << "dcp_d_" << i; 
        to_file(ss.str().c_str(), dndcp[i]);
    }
#endif
    
    Array<int> ede(edges(acc));
#ifdef DEBUG
    to_file("ede", ede);
#endif
    
    return (__compare(dndcp, ede));
}

void 
Komparator2::__decompose(const AccImage& acc, Array<float> ndcp[])
{        
    Array<std::complex<float> >* fft_a;
    fft_a= new Array<std::complex<float> >(fft(acc.A() / 
                                                acc.A().mean()).fftshift() * 
                                            __csf.A());
    for (size_t i = 0; i < 17; i++)
        ndcp[i] = ifft((*fft_a * __dcp[i]).fftshift());

    delete fft_a;    
    
    // C1
    Array<std::complex<float> >* fft_c1;
    fft_c1 = new Array<std::complex<float> >(fft(acc.C1()).fftshift() * 
                                              __csf.C1());
    for (size_t i = 17; i < 22; i++)
        ndcp[i] = ifft((*fft_c1 * __dcp[i]).fftshift());

    delete fft_c1;
        
    // C2
    Array<std::complex<float> >* fft_c2;
    fft_c2 = new Array<std::complex<float> >(fft(acc.C2()).fftshift() * 
                                              __csf.C2());
    
    for (size_t i = 22; i < 27; i++)
        ndcp[i] = ifft((*fft_c2 * __dcp[i]).fftshift());

    delete fft_c2;

    
    //////////////////////////////////////////////////////////////////////////
    //  Compute masks.
    //////////////////////////////////////////////////////////////////////////
    IntraATresholdBuilder intra_a;
    IntraCTresholdBuilder intra_c;
    InterATresholdBuilder inter_a;
    InterBTresholdBuilder inter_b;
    
    
    Array<float> es[27];
    // Compute flag.
    /*
    int temp;
    int flag_intra_a = 40 * (int) (__maskFlag / 40);
    temp = __maskFlag - flag_intra_a;
    int flag_intra_c1 = 20 * (int) (temp / 20);
    temp = temp - flag_intra_c1;
    int flag_intra_c2 = 10 * (int) (temp / 10);
    temp = temp - flag_intra_c2;
    int flag_inter_a = 4 * (int) ((module(temp, 10)) / 4);
    temp = temp - flag_inter_a;
    int flag_inter_c1 = 2 * (int) ((module(temp, 10)) / 2);
    temp = temp - flag_inter_c1;
    int flag_inter_c2 = (int) (module(temp,10));
    */
    
    // Intra A
    if (__maskIntraA)
    {
        es[0] = intra_a((ndcp[0] / (float)250), COEF1);        
        
        es[1] = intra_a((ndcp[1] / (float)250), COEF2);
        es[2] = intra_a((ndcp[2] / (float)250), COEF2);
        es[3] = intra_a((ndcp[3] / (float)250), COEF2);
        es[4] = intra_a((ndcp[4] / (float)250), COEF2);
        
        es[5] = intra_a((ndcp[5] / (float)250), COEF3);
        es[6] = intra_a((ndcp[6] / (float)250), COEF3);
        es[7] = intra_a((ndcp[7] / (float)250), COEF3);
        es[8] = intra_a((ndcp[8] / (float)250), COEF3);
        es[9] = intra_a((ndcp[9] / (float)250), COEF3);
        es[10] = intra_a((ndcp[10] / (float)250), COEF3);
        es[11] = intra_a((ndcp[11] / (float)250), COEF3);
        es[12] = intra_a((ndcp[12] / (float)250), COEF3);
        es[13] = intra_a((ndcp[13] / (float)250), COEF3);
        es[14] = intra_a((ndcp[14] / (float)250), COEF3);
        es[15] = intra_a((ndcp[15] / (float)250), COEF3);
        es[16] = intra_a((ndcp[16] / (float)250), COEF3);        
    }
    
    // Inter A
    if (__maskInterA == 4)
    {
        Array<float> tmp_cr1_B  = ndcp[17] / (float)30;
        Array<float> tmp_cr1_21 = ndcp[18] / (float)30;
        Array<float> tmp_cr1_22 = ndcp[19] / (float)30;
        Array<float> tmp_cr1_23 = ndcp[20] / (float)30;
        Array<float> tmp_cr1_24 = ndcp[21] / (float)30;
        
        Array<float> sum = tmp_cr1_21 + tmp_cr1_22 + tmp_cr1_23 + tmp_cr1_24;
        
        es[0] *= inter_b(tmp_cr1_B, COEF4, COEF5, COEF6);
        es[0] *= inter_b(sum, COEF7, COEF8, COEF9);

        Array<float> ia_tmp_cr1_B(inter_a(tmp_cr1_B, COEF10, COEF11, COEF12));
        
        es[1] *= ia_tmp_cr1_B;
        es[1] *= inter_b(tmp_cr1_21, COEF13, COEF14, COEF15);
        es[2] *= ia_tmp_cr1_B;
        es[2] *= inter_b(tmp_cr1_22, COEF13, COEF14, COEF15);
        es[3] *= ia_tmp_cr1_B;
        es[3] *= inter_b(tmp_cr1_23, COEF13, COEF14, COEF15);
        es[4] *= ia_tmp_cr1_B;
        es[4] *= inter_b(tmp_cr1_24, COEF13, COEF14, COEF15);
                
        Array<float> ib_sum = inter_a(sum, COEF19, COEF20, COEF21);
        es[5] *= ia_tmp_cr1_B;
        es[5] *= ib_sum;
        es[6] *= ia_tmp_cr1_B;
        es[6] *= ib_sum;
        es[7] *= ia_tmp_cr1_B;
        es[7] *= ib_sum;
        es[8] *= ia_tmp_cr1_B;
        es[8] *= ib_sum;
        es[9] *= ia_tmp_cr1_B;
        es[9] *= ib_sum;
        es[10] *= ia_tmp_cr1_B;
        es[10] *= ib_sum;
    }
 
    // Intra C1
    if (__maskIntraC1)
    {
        es[17] = intra_c((ndcp[17] / (float)30), COEF22, COEF23, COEF24);
        
        es[18] = intra_c((ndcp[18] / (float)30), COEF25, COEF26, COEF27);
        es[19] = intra_c((ndcp[19] / (float)30), COEF25, COEF26, COEF27);
        es[20] = intra_c((ndcp[20] / (float)30), COEF25, COEF26, COEF27);
        es[21] = intra_c((ndcp[21] / (float)30), COEF25, COEF26, COEF27);        
    }
    
    // Inter C1
    if (__maskInterC1)
    {
        Array<float> tmp_a_B = ndcp[0] / (float)250;
        Array<float> tmp_c2_B = ndcp[22] / (float)5;
        
        es[17] *= inter_a(tmp_a_B, COEF28, COEF29, COEF30);
        es[17] *= inter_a(tmp_c2_B, COEF31, COEF32, COEF33);
        
        Array<float> ia_tmp_a_B = inter_a(tmp_a_B, COEF34, COEF35, COEF36);
        Array<float> ib_tmp_c2_B = inter_b(tmp_a_B, COEF40, COEF41, COEF42);
        
        es[18] *= ia_tmp_a_B;
        es[18] *= inter_a((ndcp[1] / (float)250), COEF37, COEF38, COEF39);
        es[18] *= ib_tmp_c2_B;

        es[19] *= ia_tmp_a_B;
        es[19] *= inter_a((ndcp[2] / (float)250), COEF37, COEF38, COEF39);
        es[19] *= ib_tmp_c2_B;
        
        es[20] *= ia_tmp_a_B;
        es[20] *= inter_a((ndcp[3] / (float)250), COEF37, COEF38, COEF39);
        es[20] *= ib_tmp_c2_B;

        es[21] *= ia_tmp_a_B;
        es[21] *= inter_a((ndcp[4] / (float)250), COEF37, COEF38, COEF39);
        es[21] *= ib_tmp_c2_B;        
    }
 
    // Intra Cr2
    if (__maskIntraC2)
    {
        es[22] = intra_c((ndcp[22] / (float)5), COEF43, COEF44, COEF45);
        
        es[23] = intra_c((ndcp[23] / (float)5), COEF46, COEF47, COEF48);
        es[24] = intra_c((ndcp[24] / (float)5), COEF46, COEF47, COEF48);
        es[25] = intra_c((ndcp[25] / (float)5), COEF46, COEF47, COEF48);
        es[26] = intra_c((ndcp[26] / (float)5), COEF46, COEF47, COEF48);        
    }
    
    // Inter Cr2
    if (__maskIntraC2)
    {
        es[22] *= inter_a((ndcp[17] / (float)30), COEF49, COEF50, COEF51);
                
        es[23] *= inter_b((ndcp[18] / (float)30), COEF52, COEF53, COEF54);
        es[24] *= inter_b((ndcp[19] / (float)30), COEF52, COEF53, COEF54);
        es[25] *= inter_b((ndcp[20] / (float)30), COEF52, COEF53, COEF54);
        es[26] *= inter_b((ndcp[21] / (float)30), COEF52, COEF53, COEF54);
    }

    /////////////////////////////////////////////////////////////////////////
    //  Normalize.
    /////////////////////////////////////////////////////////////////////////    
#ifdef DEBUG
    for (size_t i = 0; i < 17; i++)
    {
        std::stringstream ss;
        ss << "es_" << i; 
        to_file(ss.str().c_str(), es[i]);
    }
#endif
        if (__maskIntraA || __maskInterA || __maskIntraC1 || __maskInterC1 || __maskIntraC2 || __maskInterC2)
            for (size_t i = 0; i < 27; i++)
                ndcp[i] /= es[i];
}

Comparison
Komparator2::__compare(Array<float> dndcp[], Array<int>& edges)
{
    //////////////////////////////////////////////////////////////////////////
    // Compute error map.
    //////////////////////////////////////////////////////////////////////////
    Array<float> emap[27];
    for (size_t i = 0; i < 27; i++)
    {
        emap[i] = Array<float>(__ndcpor[i].rows(), __ndcpor[i].cols());
        for (size_t r = 0; r < __ndcpor[i].rows(); r++)
            for (size_t c = 0; c < __ndcpor[i].cols(); c++)
                emap[i](r, c) = fabs(__ndcpor[i](r, c) - dndcp[i](r, c));
    }

#ifdef DEBUG
    for (size_t i = 0; i < 27; i++)
    {
        std::stringstream ss;
        ss << "emap_" << i; 
        to_file(ss.str().c_str(), emap[i]);
    }
#endif
    
    //////////////////////////////////////////////////////////////////////////
    // Cumul.
    //////////////////////////////////////////////////////////////////////////
    Array<float>* err_bb = new Array<float>((float)10 * emap[0] + 
                                            (float)5 * emap[17] + 
                                            emap[22]);
    Array<float>* err_21 = new Array<float>((float)10 * emap[1] + 
                                            (float)5 * emap[18] + 
                                            emap[23]);
    Array<float>* err_22 = new Array<float>((float)10 * emap[2] + 
                                            (float)5 * emap[19] + 
                                            emap[24]);
    Array<float>* err_23 = new Array<float>((float)10 * emap[3] + 
                                            (float)5 * emap[20] + 
                                            emap[25]);
    Array<float>* err_24 = new Array<float>((float)10 * emap[4] + 
                                            (float)5 * emap[21] + 
                                            emap[26]);

    Array<float>* err_2 = new Array<float>(__ndcpor[0].rows(), 
                                             __ndcpor[0].cols());
    Array<float>* err_3 = new Array<float>(__ndcpor[0].rows(),
                                             __ndcpor[0].cols());
    Array<float>* err_4 = new Array<float>(__ndcpor[0].rows(), 
                                             __ndcpor[0].cols());    
    for (size_t i = 0; i < err_2->rows(); i++) 
    {
        for (size_t j = 0; j < err_2->cols(); j++) 
        {
            (*err_2)(i, j) = (float)(powf((powf((*err_21)(i, j), 4.5) + 
                                           powf((*err_22)(i, j), 4.5) + 
                                           powf((*err_23)(i, j), 4.5) + 
                                           powf((*err_24)(i, j), 4.5)), 
                                          (1.0 / 4.5)));
            (*err_3)(i, j) = (float)(powf((powf(emap[5](i,j), 4.5) + 
                                           powf(emap[6](i,j), 4.5) + 
                                           powf(emap[7](i,j), 4.5) + 
                                           powf(emap[8](i,j), 4.5) + 
                                           powf(emap[9](i,j), 4.5) + 
                                           powf(emap[10](i,j), 4.5)), 
                                          (1.0 / 4.5)));
            (*err_4)(i, j) = (float)(powf((powf(emap[11](i,j), 4.5) + 
                                           powf(emap[12](i,j), 4.5) + 
                                           powf(emap[13](i,j), 4.5) + 
                                           powf(emap[14](i,j), 4.5) + 
                                           powf(emap[15](i,j), 4.5) + 
                                           powf(emap[16](i,j), 4.5)), 
                                          (1.0 / 4.5)));
        }
    }
    
    Array<float>* err = new Array<float>(*err_bb + (float)10 * 
                                           *err_2 + (float)10 * 
                                           *err_3 + *err_4);
    delete err_bb;
    delete err_21;
    delete err_22;
    delete err_23;
    delete err_24;
    delete err_2;
    delete err_3;
    delete err_4;
    
#ifdef DEBUG
    to_file("err", *err);
#endif
    
    // Density.
    Array<float>* errd = new Array<float>(blkproc(*err));
    
    *errd /= errd->max();
    *errd *= *errd;
    
#ifdef DEBUG
    to_file("errd", *errd);
#endif
    
    // Edges (Morpho math)
    Array<int> xored(edges.rows(), edges.cols());
    for (size_t r = 0; r < edges.rows(); r++)
    {
        for (size_t c = 0; c < edges.cols(); c++)
        {
            if((__eor(r, c) == 0) && (edges(r, c) == 0))
                xored(r, c) = 0;
            if((__eor(r, c) == 0) && (edges(r, c) == 255))
                xored(r, c) = 255;
            if((__eor(r, c) == 255) && (edges(r, c) == 255))
                xored(r, c) = 0;
            if((__eor(r, c) == 255) && (edges(r, c) == 0))
                xored(r, c) = 255;
        }
    }
    
#ifdef DEBUG
    to_file("xored", xored);
#endif

    Array<int> c11 = erode_dilate_h(xored);
#ifdef DEBUG
    to_file("c11", c11);
#endif
    Array<int> c12 = erode_dilate_v(xored);
#ifdef DEBUG
    to_file("c12", c12);
#endif
    Array<int> cored(c11.rows(), c11.cols());
    for (size_t r = 0; r < cored.rows(); r++)
        for (size_t c = 0; c < cored.cols(); c++)
            if (c11(r, c) != 0 || c12(r, c) != 0)
                cored(r, c) = 255;

#ifdef DEBUG
    to_file("cored", cored);
#endif
    Array<int> c1 = dilate_hv(cored);
#ifdef DEBUG    
    to_file("c1", c1);
#endif
    
    //////////////////////////////////////////////////////////////////////////
    // Compute Mobj.
    //////////////////////////////////////////////////////////////////////////
    Array<float> fd(errd->rows(), errd->cols());
    for (size_t r = 0; r < fd.rows(); r++)
        for (size_t c = 0; c < fd.cols(); c++)
            fd(r, c) = powf((*errd)(r, c), __alpha);
    
    Array<int> bin_c1(c1 / (int)255);

    Array<int> bin_c1_comp(c1.rows(), c1.cols());
    for (size_t r = 0; r < bin_c1.rows(); r++)
        for (size_t c = 0; c < bin_c1.cols(); c++)
            if (bin_c1(r, c) == 0)
                bin_c1_comp(r, c) = 1;
            else
                bin_c1_comp(r, c) = 0;

    // XXX : modify Array template -> Array<float> fc(bin_c1 * __co + bin_c1_comp);
    Array<float> fc(bin_c1.rows(), bin_c1.cols());
    for (size_t r = 0; r < bin_c1.rows(); r++)
        for (size_t c = 0; c < bin_c1.cols(); c++)
            fc(r, c) = (float)bin_c1(r, c) * __co + (float)bin_c1_comp(r, c);

    Array<float> ft = fc * fd;
    
    Array<float> tmp(ft.rows(), ft.cols());
    for (size_t r = 0; r < ft.rows(); r++)
        for (size_t c = 0; c < ft.cols(); c++)
            tmp(r, c) = powf(ft(r, c) * (*err)(r, c), __beta);
    
    float mobj = powf(tmp.mean(), 1.0 / __beta);
    
    float nobj = -1;
    
    if (mobj < ((__b - 5) / __a)) 
        nobj = 5;
    else if ((((__b - 1) / __a) >= mobj) && (mobj > ((__b - 5) / __a))) 
        nobj = (-(__a) * mobj) + __b;
    else if (mobj > ((__b - 5) / __a)) 
        nobj = 1;

    return Comparison(mobj, nobj, *err);
}


//////////////////////////////////////////////////////////////////////////////
// Helpers functions
//////////////////////////////////////////////////////////////////////////////

int
module(int i, int j)
{
    if(j == 0)
    {
        return i;
    }
    else
    {
        int temp = (int) (i/j);
        if((i - (j * temp)) >= 0)
            return (i - (j * temp));
        else
            return (j + (i - (j * temp)));
    }
}

#define W_SIZE 32
Array<float> blkproc(Array<float>& in)
{    
    size_t nbr = in.rows() / W_SIZE;
    size_t nbc = in.cols() / W_SIZE;
    
    if (nbr * W_SIZE != in.rows())
        nbr++;
    
    if (nbc * W_SIZE != in.cols())
        nbc++;
        
    Array<float> out(in.rows(), in.cols());
    
    for (size_t rb = 0; rb < nbr; rb++)
    {
        for (size_t cb = 0; cb < nbc; cb++)
        {
            Array<float> marray(W_SIZE, W_SIZE);
            size_t r;
            size_t c;
            size_t rm;
            size_t cm;
            
            for (r = rb * W_SIZE, rm = 0; r < (rb + 1) * W_SIZE; r++, rm++)
                for (c = cb * W_SIZE, cm = 0; c < (cb + 1) * W_SIZE; c++, cm++)
                    marray(rm, cm) = in(r, c);
            
            float m = marray.mean();
            
            for (r = rb * W_SIZE; r < (rb + 1) * W_SIZE; r++)
                for (c = cb * W_SIZE; c < (cb + 1) * W_SIZE; c++)
                    out(r, c) = m;
        }
    }
        
    return out;
}

#define SIGMA 1.0
#define THIGH 25.0
#define TLOW 6.0
Array<int>
edges(const AccImage& acc)
{
    Array<int> out(acc.rows(), acc.cols());
    CannyBuilder cb(SIGMA, TLOW, THIGH);
    
    Array<int> cor_a = cb(acc.A() / acc.A().mean());
    Array<int> cor_c1 = cb(acc.C1() / acc.C1().mean());
    Array<int> cor_c2 = cb(acc.C2() / acc.C2().mean());
    
    // cor_a | cor_c1
    for (size_t r = 0; r < acc.rows(); r++)
        for (size_t c = 0; c < acc.cols(); c++)
            if (cor_a(r, c) == 255 || cor_c1(r, c) == 255)
                out(r, c) = 255;
    
    // cor_a | cor_c1 | cor_c2
    for (size_t r = 0; r < acc.rows(); r++)
        for (size_t c = 0; c < acc.cols(); c++)
            if (out(r, c) == 255 || cor_c2(r, c) == 255)
                out(r, c) = 255;
    
    return out;
}

int min(int a, int b)
{
    return a < b ? a : b;
}

Array<int> 
erode_dilate_h(Array<int>& in)
{
    int  pdisth1;
    int  pdisth2;
    
    Array<int> erode(in.rows(), in.cols());
    Array<int> dilate(in.rows(), in.cols());
    
    // Horizontal
    for (size_t r = 0; r < in.rows(); r++)
    {
        size_t cd = 0;
        pdisth1 = 0;
        for (size_t c = 0; c < in.cols(); c++)
        {
            pdisth1 = in(r, c) > 0 ? min(pdisth1 + 1, 4) : 0;
            erode(r, c) = pdisth1 == 4 ? 255 : 0;
        }
        
        pdisth2 = 4;
        for (size_t c = 0; c < in.cols() + 3; c++)
        {
            if (c < in.cols())
                pdisth2 = erode(r, c) == 0 ? min(pdisth2 + 1, 4) : 0;

            if (c >= 3)
                dilate(r, cd++) = pdisth2 < 4 ? 255 : 0;
        }
    }

    return dilate;
}

Array<int> 
erode_dilate_v(Array<int>& in)
{
    int  pdisth1;
    int  pdisth2;
    
    Array<int> erode(in.rows(), in.cols());
    Array<int> dilate(in.rows(), in.cols());
    
    // Vertical
    for (size_t c = 0; c < in.cols(); c++)
    {
        size_t rd = 0;
        pdisth1 = 0;
        for (size_t r = 0; r < in.rows(); r++)
        {
            pdisth1 = in(r, c) > 0 ? min(pdisth1 + 1, 4) : 0;
            erode(r, c) = pdisth1 == 4 ? 255 : 0;
        }
        
        pdisth2 = 4;
        for (size_t r = 0; r < in.rows() + 3; r++)
        {
            if (r < in.rows())
                pdisth2 = erode(r, c) == 0 ? min(pdisth2 + 1, 4) : 0;
            
            if (r >= 3)
                dilate(rd++, c) = pdisth2 < 4 ? 255 : 0;
        }
    }
    
    return dilate;
}
    
Array<int> 
dilate_hv(Array<int>& in)
{
    int  pdist;
    
    Array<int> dilate1(in.rows(), in.cols());
    Array<int> dilate2(in.rows(), in.cols());
    
    // Horizontal
    for (size_t r = 0; r < in.rows(); r++)
    {
        size_t cd = 0;
        pdist = 1;
        for (size_t c = 0; c < in.cols() + 1; c++)
        {
            if (c < in.cols())
                pdist = in(r, c) == 0 ? min(pdist + 1, 3) : 0;
            
            if (c >= 1)
                dilate1(r, cd++) = pdist < 3 ? 255 : 0;
        }
    }
    
    // Vertical
    for (size_t c = 0; c < in.cols(); c++)
    {
        size_t rd = 0;
        pdist = 1;
        for (size_t r = 0; r < in.rows() + 1; r++)
        {
            if (r < in.rows())
                pdist = dilate1(r, c) == 0 ? min(pdist + 1, 3) : 0;
            
            if (r >= 1)
                dilate2(rd++, c) = pdist < 3 ? 255 : 0;
        }
    }
    
    return dilate2;
}

Array<std::complex<float> > fft(const Array<float>& in)
{
    Complex* ic = new Complex[in.rows()*in.cols()];
    Complex* oc = new Complex[in.rows()*in.cols()];
    
    for (size_t r = 0; r < in.rows(); r++)
        for (size_t c = 0; c < in.cols(); c++)
            ic[r * in.cols() + c] = Complex((double)in(r, c), 0.0);
    
    fft2d forward(in.rows(), in.cols(), FFTW_FORWARD, ic, oc);
    forward.fft(ic, oc);
    
    Array<std::complex<float> > out(in.rows(), in.cols());
    for (size_t r = 0; r < in.rows(); r++)
        for (size_t c = 0; c < in.cols(); c++)
            out(r, c) = oc[r * in.cols() + c];
    
    delete [] ic;
    delete [] oc;
    
    return out;
}

Array<float> ifft(const Array<std::complex<float> >& in)
{
    Complex* ic = new Complex[in.rows()*in.cols()];
    Complex* oc = new Complex[in.rows()*in.cols()];
    
    for (size_t r = 0; r < in.rows(); r++)
        for (size_t c = 0; c < in.cols(); c++)
            ic[r * in.cols() + c] = in(r, c);

    fft2d backward(in.rows(), in.cols(), FFTW_BACKWARD, ic, oc);
    backward.fft0Normalized(ic, oc);

    Array<float> out(in.rows(), in.cols());
    for (size_t r = 0; r < in.rows(); r++)
        for (size_t c = 0; c < in.cols(); c++)
            out(r, c) = (float)oc[r * in.cols() + c].real();
    
    delete [] ic;
    delete [] oc;
    
    return out;
}

template<class T>
void to_file(const char* fname, const Array<T>& array)
{
    std::stringstream ss;
    ss << fname << ".dat";
    std::ofstream os(ss.str().c_str());
    os << array;
    os.close();
}
