#include <QtGui>

#include "Komparator2.h"
#include "RgbImage.h"
#include "Comparison.h"
#include "QKomparatorMainWindow.h"
#include "QKomparatorErrorMapDialog.h"

QKomparatorMainWindow::QKomparatorMainWindow()
{
    ui.setupUi(this);
    
    // Parameters
    ui.lineEditAlpha->setValidator(new QDoubleValidator(ui.lineEditAlpha));
    ui.lineEditBeta->setValidator(new QDoubleValidator(ui.lineEditBeta));
    ui.lineEditCo->setValidator(new QDoubleValidator(ui.lineEditCo));
    ui.lineEditA->setValidator(new QDoubleValidator(ui.lineEditA));
    ui.lineEditB->setValidator(new QDoubleValidator(ui.lineEditB));
    
    // Results.
    ui.tableWidget->horizontalHeader()->setResizeMode(0, QHeaderView::Stretch);
    QStringList hHeader;
    hHeader << QString("File name") << QString("Measure") << QString("Score");
    ui.tableWidget->setHorizontalHeaderLabels(hHeader);
    ui.tableWidget->verticalHeader()->hide();
    ui.progressBar->hide();
    
    // Bug
    ui.checkBoxInterA->setEnabled(false);
    ui.checkBoxInterC1->setEnabled(false);
    ui.checkBoxInterC2->setEnabled(false);
    
    setupDefault();
}

void
QKomparatorMainWindow::setupDefault()
{
    ui.lineEditAlpha->setText("1");
    ui.lineEditBeta->setText("4");
    ui.lineEditCo->setText("5");
    ui.lineEditA->setText("0.000751");
    ui.lineEditB->setText("3.8");
    ui.checkBoxIntraA->setChecked(true);
    ui.checkBoxIntraC1->setChecked(true);
    ui.checkBoxIntraC2->setChecked(true);
    
}

void
QKomparatorMainWindow::on_actionNew_triggered(){
    // load files names
    original = QFileDialog::getOpenFileName(
                    this,
                    "Choose original file",
                    "/home",
                    "Images (*.ppm *.pbm *.ppm)");
    
    if (original.isEmpty())
        return;
                        
    degraded = QFileDialog::getOpenFileNames(
                    this,
                    "Choose degraded files",
                    "/home",
                    "Images (*.ppm *.pbm *.ppm)");
    

    if (degraded.empty())
        return;
        
    setWindowTitle(original);
    
    // compare  
    RgbImage originalRgb(original.toStdString());
    
    float alpha = ui.lineEditAlpha->text().toFloat();
    float beta = ui.lineEditBeta->text().toFloat();
    float co = ui.lineEditCo->text().toFloat();
    float a = ui.lineEditA->text().toFloat();
    float b = ui.lineEditB->text().toFloat();
    bool mask_intra_a = ui.checkBoxIntraA->isChecked();
    bool mask_inter_a = ui.checkBoxInterA->isChecked();
    bool mask_intra_c1 = ui.checkBoxIntraC1->isChecked();
    bool mask_inter_c1 = ui.checkBoxInterC1->isChecked();
    bool mask_intra_c2 = ui.checkBoxIntraC2->isChecked();
    bool mask_inter_c2 = ui.checkBoxInterC2->isChecked();
    
    Komparator2 komparator(originalRgb, alpha, beta, co, a, b, mask_intra_a, mask_inter_a,
        mask_intra_c1, mask_inter_c1, mask_intra_c2, mask_inter_c2);

    ui.progressBar->setRange(0, degraded.size());
    ui.progressBar->show();
    // compute
    comparisons.clear();
    ui.tableWidget->setRowCount(0);
    for (int i = 0; i < degraded.size(); i ++)
    {
        ui.progressBar->setValue(i);
        qApp->processEvents();
         
        RgbImage degradedRgb(degraded[i].toStdString());
        Comparison comparison = komparator(degradedRgb);
        comparisons.push_back(comparison);
        
        int row = ui.tableWidget->rowCount();
        ui.tableWidget->insertRow(row);
        
        ui.tableWidget->setItem(row, 0, new QTableWidgetItem(degraded[i]));
        
        QString mobj;
        mobj.setNum(comparison.mobj());
        ui.tableWidget->setItem(row, 1, new QTableWidgetItem(mobj));
        
        QString nobj;
        nobj.setNum(comparison.nobj());
        ui.tableWidget->setItem(row, 2, new QTableWidgetItem(nobj));
    }
    ui.progressBar->hide();
}

void
QKomparatorMainWindow::on_actionSave_triggered()
{
    QString dirname = QFileDialog::getExistingDirectory(this, tr("Save directory"));
    if (dirname.isEmpty())
        return;
        
    QDir dir(dirname);
    
    QFileInfo o_info(original);
    
    // Save results.
    QFile file(dir.absoluteFilePath(o_info.baseName() + "_results.txt"));
    if (!file.open(QIODevice::ReadWrite | QIODevice::Text))
    {
        QErrorMessage error(this);
        error.showMessage("Can't open file");
        return;
    }
    
    file.setTextModeEnabled(true);
    QTextStream stream(&file);
    
    for (int i = 0; i < degraded.size(); i ++)
    {
        stream << degraded[i] << " " << comparisons[i].mobj() << " " << comparisons[i].nobj() << endl;
    }
    stream.flush();
    
    // Save error maps.
    for (int i = 0; i < degraded.size(); i ++)
    {
        QFileInfo d_info(degraded[i]);
        QFile file_map(dir.absoluteFilePath(o_info.baseName() + "_" + d_info.baseName() + "_emap.txt"));
        if (!file_map.open(QIODevice::ReadWrite | QIODevice::Text))
        {
            QErrorMessage error(this);
            error.showMessage("Can't open file");
            return;
        }
        
        QTextStream stream_map(&file_map);
        for (size_t r = 0; r < comparisons[i].err().rows(); r ++)
        {
            for (size_t c = 0; c < comparisons[i].err().cols(); c ++)
            {
                stream_map << comparisons[i].err()(r, c) << "\t";
            }
            stream_map << endl;
        }
        stream_map.flush();
    }
}

void
QKomparatorMainWindow::on_tableWidget_doubleClicked(const QModelIndex& index)
{
    QFileInfo o_info(original);
    QFileInfo d_info(degraded[index.row()]);
    QKomparatorErrorMapDialog* map = new QKomparatorErrorMapDialog(o_info.completeBaseName(), d_info.completeBaseName(), comparisons[index.row()].err());
    map->show();
}
