extern "C"
{
#include <ppm.h>
}

#include "RgbImage.h"

RgbImage::RgbImage()
{
    __rows = 0;
    __cols = 0;
}

RgbImage::RgbImage(const std::string fname)
{
    // Read ppm file.
    FILE* in = pm_openr(fname.c_str());
    int cols;
    int rows;
    pixval maxval;
    pixel** pixels = ppm_readppm(in, &cols, &rows, &maxval);
    pm_close(in);
    
    __rows = (size_t)rows;
    __cols = (size_t)cols;
    __R = Array<unsigned int>(rows, cols);
    __G = Array<unsigned int>(rows, cols);
    __B = Array<unsigned int>(rows, cols);
    
    for (size_t r = 0; r < (size_t)rows; r++)
    {
        for (size_t c = 0; c < (size_t)cols; c ++)
        {
            __R(r, c) = (unsigned int)PPM_GETR(pixels[r][c]);
            __G(r, c) = (unsigned int)PPM_GETG(pixels[r][c]);
            __B(r, c) = (unsigned int)PPM_GETB(pixels[r][c]);
        }
    }    
}

RgbImage::~RgbImage()
{
}

size_t
RgbImage::rows() const
{
    return __rows;
}

size_t
RgbImage::cols() const
{
    return __cols;
}

const Array<unsigned int>&
RgbImage::R() const
{
    return __R;
}

const Array<unsigned int>&
RgbImage::G() const
{
    return __G;
}

const Array<unsigned int>&
RgbImage::B() const
{
    return __B;
}

std::ostream& 
operator<<(std::ostream& os, const RgbImage& img)
{
    os << "R" << std::endl;
    os << img.__R << std::endl;
    os << "G" << std::endl;
    os << img.__G << std::endl;
    os << "B" << std::endl;
    os << img.__B << std::endl;
    
    return os;
}