#include <vector>

#include "RgbImage.h"
#include "Komparator.h"
#include "Metric.h"
#include "ArgumentLoader.h"

/*
#include "CannyBuilder.h"
//#include <fstream>
//#include <sstream>

int main(int argc, char*argv[])
{

    RgbImage orgb(argv[1]);
    AccImage acc(orgb);
    CannyBuilder cb(1.0, 25.0, 6.0);
    cb(acc.A());
    return 0;
}
*/

int nr_args = 5;
const char* args[][2] = {
{"h", "Print this message."},
{"of", "Original image file."},
{"df", "Degraded image(s) file(s)."},
{"nobj", "Compute objective note."},
{"mask", "Set mask flag."}
};

#define HELPOPT args[0][0]
#define ORIGOPT args[1][0]
#define DEGROPT args[2][0]
#define NOBJOPT args[3][0]
#define MASKOPT args[4][0]

void usage()
{
    std::cout << "komparator [-h] -o <file> -d <file ...> [-nobj <alpha> <beta> <co> <a> <b>] [-mask <flag>]" << std::endl;
    for (int i = 0; i < nr_args; i++)
        std::cout << "\t-" << args[i][0] << "\t" << args[i][1] << std::endl;
};

int main(int argc, char*argv[])
{
    std::string original_fname;
    std::vector<std::string> degraded_fname;
    bool compute_nobj = false;
    float alpha = 1;
    float beta = 4;
    float co = 5;
    float a = 0.000751;
    float b = 3.8;
    int mask = 70;
        
    ArgumentLoader loader(argc, argv);
    loader.addPattern(HELPOPT, 0, false);
    loader.addPattern(ORIGOPT, 1, true);
    loader.addPattern(DEGROPT, -1, true);
    loader.addPattern(NOBJOPT, 5, false);
    loader.addPattern(MASKOPT, 1, false);
    
    if (loader.parseArguments())
    {
        // Help wanted.
        if (loader.getArgument(HELPOPT).isInitiated())
        {
            usage();
            return 0;
        }
        
        original_fname = loader.getArgument(ORIGOPT).getSubArg(0);
        
        unsigned int nr_deg = loader.getArgument(DEGROPT).getSubArgCount();
        for (unsigned int i = 0; i < nr_deg; i++)
            degraded_fname.push_back(loader.getArgument(DEGROPT).getSubArg(i));
        
        Argument nobj_arg = loader.getArgument(NOBJOPT);
        if (nobj_arg.isInitiated())
        {
            compute_nobj = true;
            alpha = float(nobj_arg.getSubArgFloat(0));
            beta = float(nobj_arg.getSubArgFloat(1));
            co = float(nobj_arg.getSubArgFloat(2));
            a = float(nobj_arg.getSubArgFloat(3));
            b = float(nobj_arg.getSubArgFloat(4));
        }
        
        Argument mask_arg = loader.getArgument(MASKOPT);
        if (mask_arg.isInitiated())
        {
            mask = mask_arg.getSubArgInt(0);
        }
            
    }
    else
    {
        std::vector < string > errors = loader.getErrors();
        for (unsigned int i = 0; i < errors.size(); i++)
        {
            std::cout << errors[i] << std::endl;
        }
        usage();
        return -1;
    }
    
    RgbImage orgb(original_fname);
    Komparator komparator(orgb, alpha, beta, co, a, b, mask, compute_nobj);
    
    for (size_t i = 0; i < degraded_fname.size(); i++)
    {
        RgbImage drgb(degraded_fname[i]);
        
        Metric m = komparator(drgb);
        
        std::cout << "Mobj : " << m.mobj() << std::endl;
        if (compute_nobj)
            std::cout << "Nobj : " << m.nobj() << std::endl;
    }
    
    return 0;
}

